import unittest
from enum import Enum, auto

from fsmate import ImpossibleTransitionError, StateDescriptor
from fsmate._state import AttributeStateStorage, StateDispatcher
from unittest.mock import MagicMock, call


class State(Enum):
    A = auto()
    B = auto()
    C = auto()


class WrongState(Enum):
    D = auto()
    E = auto()
    F = auto()


class TestStateAttribute(unittest.TestCase):
    def setUp(self) -> None:
        class Stub:
            state = StateDescriptor(State, State.A)

        self.obj = Stub()

    def test_initital_state(self):
        self.assertEqual(self.obj.state, State.A)

    def test_forbid_state_direct_change(self):
        with self.assertRaisesRegex(
            AttributeError, 'Cannot change state directly. Use transitions'
        ):
            self.obj.state = State.B

    def test_default_attribute_change(self):
        with self.assertRaises(AttributeError):
            self.obj._state

        self.obj._state = State.B
        self.assertEqual(self.obj.state, State.B)

    def test_custom_atribute_change(self):
        class Stub:
            state_attribute: State = State.A
            state = StateDescriptor(State, state_storage=AttributeStateStorage('state_attribute'))

        obj = Stub()

        self.assertEqual(obj.state, State.A)

        obj.state_attribute = State.B
        self.assertEqual(obj.state, State.B)


class TestDeclareTransitions(unittest.TestCase):
    def test_destination_state_not_found(self):
        with self.assertRaisesRegex(ValueError, 'Destination state not found'):

            class _:
                state = StateDescriptor(State, State.A)
                to_d = state.transition(State.A, WrongState.D)

    def test_source_state_not_found(self):
        with self.assertRaisesRegex(ValueError, 'Source state not found'):

            class _:
                state = StateDescriptor(State, State.A)
                from_d_to_a = state.transition(WrongState.D, State.A)

    def test_one_of_sources_not_found(self):
        with self.assertRaisesRegex(ValueError, 'Source state not found'):

            class _:
                state = StateDescriptor(State, State.A)
                from_b_or_d_to_a = state.transition([State.B, WrongState.D], State.A)


class TestTransitions(unittest.TestCase):
    def setUp(self) -> None:
        class Stub:
            state = StateDescriptor(State, State.A)

            to_b = state.transition(State.A, State.B)
            to_c = state.transition(State.B, State.C)
            to_a_from_c = state.transition(State.C, State.A)
            to_b_from_a_or_c = state.transition([State.A, State.C], State.B)

        self.obj = Stub()

    def test_valid_transition(self):
        self.assertEqual(self.obj.state, State.A)
        self.obj.to_b()
        self.assertEqual(self.obj.state, State.B)

        self.obj.to_c()
        self.assertEqual(self.obj.state, State.C)

        self.obj.to_a_from_c()
        self.assertEqual(self.obj.state, State.A)

        self.obj.to_b_from_a_or_c()
        self.assertEqual(self.obj.state, State.B)

    def test_invalid_transition(self):
        with self.assertRaises(ImpossibleTransitionError):
            self.obj.to_c()

    def test_multiple_sources(self):
        self.obj.to_b_from_a_or_c()
        self.assertEqual(self.obj.state, State.B)

        self.obj._state = State.C
        self.obj.to_b_from_a_or_c()
        self.assertEqual(self.obj.state, State.B)

        self.obj._state = State.B
        with self.assertRaises(ImpossibleTransitionError):
            self.obj.to_b_from_a_or_c()


class TestMethodOverload(unittest.TestCase):
    def test_state_dispatcher(self):
        class StubStateStorage:
            def get_state(self, instance):
                return self.state

            def set_state(self, instance, state):
                self.state = state

        def fallback(x):
            return x

        def b_func(x):
            return x * 2

        storage = StubStateStorage()
        storage.state = State.A
        dispatcher = StateDispatcher(storage, State, fallback)
        dispatcher.register(b_func, State.B)

        self.assertEqual(dispatcher.dispatch(None, 10), 10)
        storage.set_state(None, State.B)
        self.assertEqual(dispatcher.dispatch(None, 10), 20)
        storage.set_state(None, State.C)
        self.assertEqual(dispatcher.dispatch(None, 10), 10)

        with self.assertRaisesRegex(ValueError, 'Target state not found'):
            dispatcher.register(lambda x: x, WrongState.D)

        with self.assertRaisesRegex(ValueError, 'Function is already overloaded for state'):
            dispatcher.register(lambda x: x, State.B)

    def test_overload(self):
        class Stub:
            state = StateDescriptor(State, State.A)

            to_b = state.transition(State.A, State.B)
            to_c = state.transition(State.B, State.C)

            @state.dispatch
            def foo(self):
                return 0

            @foo.overload(State.B)
            def _(self):
                return 1

        obj = Stub()

        self.assertEqual(obj.foo(), 0)

        obj.to_b()
        self.assertEqual(obj.foo(), 1)

        obj.to_c()
        self.assertEqual(obj.foo(), 0)


class TestCallbacks(unittest.TestCase):
    def test_transition_callbacks(self) -> None:
        callback_mock = MagicMock()

        class Stub:
            state = StateDescriptor(State, State.A)

            to_b = state.transition(State.A, State.B)
            to_c = state.transition(State.B, State.C)
            to_a = state.transition([State.B, State.C], State.A)

            @state.on_transition(to_b)  # type: ignore
            def on_to_b(self, from_state: State, to_state: State) -> None:
                callback_mock('to_b', self, from_state, to_state)

            @state.on_transition(to_c)  # type: ignore
            def on_to_c(self, from_state: State, to_state: State) -> None:
                callback_mock('to_c', self, from_state, to_state)

            @state.on_transition(to_a)  # type: ignore
            def on_to_a(self, from_state: State, to_state: State) -> None:
                callback_mock('to_a', self, from_state, to_state)

            @state.on_transition(to_b, to_c)  # type: ignore
            def on_to_b_to_c(self, from_state: State, to_state: State):
                callback_mock('to_b_to_c', self, from_state, to_state)

            @state.on_transition  # type: ignore
            def on_to_any(self, from_state: State, to_state: State) -> None:
                callback_mock('to_any', self, from_state, to_state)

        obj = Stub()

        obj.to_b()
        callback_mock.assert_has_calls(
            [
                call('to_b', obj, State.A, State.B),
                call('to_b_to_c', obj, State.A, State.B),
                call('to_any', obj, State.A, State.B),
            ]
        )
        callback_mock.reset_mock()

        obj.to_c()
        callback_mock.assert_has_calls(
            [
                call('to_c', obj, State.B, State.C),
                call('to_b_to_c', obj, State.B, State.C),
                call('to_any', obj, State.B, State.C),
            ]
        )
        callback_mock.reset_mock()

        obj.to_a()
        callback_mock.assert_has_calls(
            [
                call('to_a', obj, State.C, State.A),
                call('to_any', obj, State.C, State.A),
            ]
        )
        callback_mock.reset_mock()

        obj.to_b()
        callback_mock.assert_has_calls(
            [
                call('to_b', obj, State.A, State.B),
                call('to_b_to_c', obj, State.A, State.B),
                call('to_any', obj, State.A, State.B),
            ]
        )
        callback_mock.reset_mock()

        obj.to_a()
        callback_mock.assert_has_calls(
            [
                call('to_a', obj, State.B, State.A),
                call('to_any', obj, State.B, State.A),
            ]
        )

    def test_enter_state_callbaks(self):
        callback_mock = MagicMock()

        class Stub:
            state = StateDescriptor(State, State.A)

            to_b = state.transition(State.A, State.B)
            to_c = state.transition(State.B, State.C)
            to_a = state.transition([State.B, State.C], State.A)

            @state.on_state_entered(State.A)  # type: ignore
            def on_a_entered(self, from_state: State, to_state: State):
                callback_mock('A_entered', self, from_state, to_state)

            @state.on_state_entered(State.B)  # type: ignore
            def on_b_entered(self, from_state: State, to_state: State):
                callback_mock('B_entered', self, from_state, to_state)

            @state.on_state_entered(State.C)  # type: ignore
            def on_c_entered(self, from_state: State, to_state: State):
                callback_mock('C_entered', self, from_state, to_state)

            @state.on_state_entered  # type: ignore
            def on_any_entered(self, from_state: State, to_state: State):
                callback_mock('any_entered', self, from_state, to_state)

            @state.on_state_entered(State.A, State.B)  # type: ignore
            def on_a_b_entered(self, from_state: State, to_state: State):
                callback_mock('AB_entered', self, from_state, to_state)

        obj = Stub()

        obj.to_b()
        callback_mock.assert_has_calls(
            [
                call('any_entered', obj, State.A, State.B),
                call('B_entered', obj, State.A, State.B),
                call('AB_entered', obj, State.A, State.B),
            ],
            any_order=True,
        )
        callback_mock.reset_mock()

        obj.to_c()
        callback_mock.assert_has_calls(
            [
                call('any_entered', obj, State.B, State.C),
                call('C_entered', obj, State.B, State.C),
            ],
            any_order=True,
        )
        callback_mock.reset_mock()

        obj.to_a()
        callback_mock.assert_has_calls(
            [
                call('any_entered', obj, State.C, State.A),
                call('AB_entered', obj, State.C, State.A),
                call('A_entered', obj, State.C, State.A),
            ],
            any_order=True,
        )
        callback_mock.reset_mock()

        obj.to_b()
        callback_mock.assert_has_calls(
            [
                call('any_entered', obj, State.A, State.B),
                call('B_entered', obj, State.A, State.B),
                call('AB_entered', obj, State.A, State.B),
            ],
            any_order=True,
        )
        callback_mock.reset_mock()

        obj.to_a()
        callback_mock.assert_has_calls(
            [
                call('AB_entered', obj, State.B, State.A),
                call('A_entered', obj, State.B, State.A),
                call('any_entered', obj, State.B, State.A),
            ],
            any_order=True,
        )

    def test_exit_state_callbaks(self):
        callback_mock = MagicMock()

        class Stub:
            state = StateDescriptor(State, State.A)

            to_b = state.transition(State.A, State.B)
            to_c = state.transition(State.B, State.C)
            to_a = state.transition([State.B, State.C], State.A)

            @state.on_state_exited(State.A)  # type: ignore
            def on_a_exited(self, from_state: State, to_state: State):
                callback_mock('A_exited', self, from_state, to_state)

            @state.on_state_exited(State.B)  # type: ignore
            def on_b_exited(self, from_state: State, to_state: State):
                callback_mock('B_exited', self, from_state, to_state)

            @state.on_state_exited(State.C)  # type: ignore
            def on_c_exited(self, from_state: State, to_state: State):
                callback_mock('C_exited', self, from_state, to_state)

            @state.on_state_exited  # type: ignore
            def on_any_exited(self, from_state: State, to_state: State):
                callback_mock('any_exited', self, from_state, to_state)

            @state.on_state_exited(State.A, State.B)  # type: ignore
            def on_a_b_exited(self, from_state: State, to_state: State):
                callback_mock('AB_exited', self, from_state, to_state)

        obj = Stub()

        obj.to_b()
        callback_mock.assert_has_calls(
            [
                call('any_exited', obj, State.A, State.B),
                call('A_exited', obj, State.A, State.B),
                call('AB_exited', obj, State.A, State.B),
            ],
            any_order=True,
        )
        callback_mock.reset_mock()

        obj.to_c()
        callback_mock.assert_has_calls(
            [
                call('any_exited', obj, State.B, State.C),
                call('B_exited', obj, State.B, State.C),
            ],
            any_order=True,
        )
        callback_mock.reset_mock()

        obj.to_a()
        callback_mock.assert_has_calls(
            [
                call('any_exited', obj, State.C, State.A),
                call('C_exited', obj, State.C, State.A),
            ],
            any_order=True,
        )
        callback_mock.reset_mock()

        obj.to_b()
        callback_mock.assert_has_calls(
            [
                call('any_exited', obj, State.A, State.B),
                call('A_exited', obj, State.A, State.B),
                call('AB_exited', obj, State.A, State.B),
            ],
            any_order=True,
        )
        callback_mock.reset_mock()

        obj.to_a()
        callback_mock.assert_has_calls(
            [
                call('AB_exited', obj, State.B, State.A),
                call('B_exited', obj, State.B, State.A),
                call('any_exited', obj, State.B, State.A),
            ],
            any_order=True,
        )



from enum import Enum, auto

from fsmate import StateDescriptor
from fsmate._state import ImpossibleTransitionError


class State(Enum):
    Username = auto()
    Password = auto()
    AccessGranted = auto()
    AccessDenied = auto()


class SigninPage:
    state = StateDescriptor(
        State,
        initial_state=State.Username,
    )

    # username is valid
    confirm_username = state.transition(
        State.Username,
        State.Password,
    )

    # username not found
    reject_username = state.transition(
        State.Username,
        State.AccessDenied,
    )

    # authenticated
    confirm_password = state.transition(
        State.Password,
        State.AccessGranted,
    )

    # wrong password - show error screen
    reject_password = state.transition(
        State.Password,
        State.AccessDenied,
    )

    # try again
    retry = state.transition(
        State.AccessDenied,
        State.Username,
    )


page = SigninPage()
print(page.state)  # State.Username

# Enter wrong username
page.reject_username()
print(page.state)  # State.AccessDenied

page.retry()

print(page.state)  # State.Username
page.confirm_username()
page.confirm_password()
print(page.state)  # State.AccessGranted

try:
    page.retry()
except ImpossibleTransitionError as error:
    print(error)

try:
    page.state = State.Username
except AttributeError as error:
    print(error)



from __future__ import annotations

from collections import defaultdict
from collections.abc import Collection
from enum import Enum
from typing import Any, Callable, Generic, NoReturn, Optional, Protocol, TypeVar, Union
from typing_extensions import ParamSpec


class ImpossibleTransitionError(Exception):
    pass


class StateStorage(Protocol):
    def get_state(self, instance: object) -> Enum:
        raise NotImplementedError

    def set_state(self, instance: object, state: Enum) -> None:
        raise NotImplementedError


class AttributeStateStorage:
    def __init__(self, attr_name: str) -> None:
        self._attr_name = attr_name

    def get_state(self, instance: object) -> Enum:
        return getattr(instance, self._attr_name)  # type: ignore[no-any-return]

    def set_state(self, instance: object, state: Enum) -> None:
        return setattr(instance, self._attr_name, state)


class ProxyStateStorage:
    def __init__(
        self, getter: Callable[[object], Enum], setter: Callable[[object, Enum], None]
    ) -> None:
        self.get_state = getter
        self.set_state = setter


class StateTransition:
    """
    Retrieves state of the object and checks if the transition is possible.
    Updates state if so.
    Raises ImpossibleTransitionError otherwise
    """

    def __init__(self, source: Collection[Enum], dest: Enum, state_storage: StateStorage) -> None:
        self._source = source
        self._dest = dest
        self._storage = state_storage

        self._callbacks: list[Callable] = []

    def __get__(
        self, instance: object, objtype: type
    ) -> Union['StateTransition', Callable[[], None]]:
        if instance is None:
            return self

        def _transition() -> None:
            state = self._storage.get_state(instance)
            if state not in self._source:
                raise ImpossibleTransitionError()

            self._storage.set_state(instance, self._dest)
            for callback in self._callbacks:
                callback(instance, state, self._dest)

        return _transition

    def __call__(self) -> None:
        pass

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(source={self._source!r}, dest={self._dest!r}, state_storage={self._storage!r})'

    def _register_callback(self, func: Callable) -> None:
        self._callbacks.append(func)


T = TypeVar('T')
P = ParamSpec('P')


class StateDispatcher(Generic[P, T]):
    def __init__(
        self, state_storage: StateStorage, all_states: type[Enum], fallback: Callable[P, T]
    ) -> None:
        self._all_states = all_states
        self._fallback = fallback
        self._state_storage = state_storage
        self._dispatch_table: dict[Enum, Callable[P, T]] = {}

    def register(self, func: Callable[P, T], *states: Enum) -> None:
        for state in states:
            if state not in self._all_states:
                raise ValueError('Target state not found', state)

            if state in self._dispatch_table:
                raise ValueError('Function is already overloaded for state', state)

            self._dispatch_table[state] = func

    def _dispatch(self, state: Enum) -> Callable[P, T]:
        return self._dispatch_table.get(state, self._fallback)

    def dispatch(self, instance: object, *args: P.args, **kwargs: P.kwargs) -> T:
        current_state = self._state_storage.get_state(instance)
        func = self._dispatch(current_state)

        return func(*args, **kwargs)


class StateDispatchedMethod(Generic[P, T]):
    def __init__(self, dispatcher: StateDispatcher[P, T]) -> None:
        self._dispatcher = dispatcher

    def __get__(self, instance: object, objtype: type) -> Callable[P, T]:
        if instance is None:
            return self  # type: ignore[return-value]

        def dispatched_method(*args: P.args, **kwargs: P.kwargs) -> T:
            return self._dispatcher.dispatch(instance, instance, *args, **kwargs)  # type: ignore[arg-type]

        return dispatched_method

    def overload(self, *states: Enum) -> Callable[[Callable[P, T]], 'StateDispatchedMethod[P, T]']:
        def deco(meth: Callable[P, T]) -> 'StateDispatchedMethod[P, T]':
            self._dispatcher.register(meth, *states)
            return self

        return deco


class StateDescriptor:
    def __init__(
        self,
        states: type[Enum],
        initial_state: Optional[Enum] = None,
        state_storage: Optional[StateStorage] = None,
    ):
        if initial_state is None == state_storage is None:
            raise ValueError('Expected only one: initial_state or state_storage')

        self._all_states = states
        self._initial_state = initial_state
        self._state_storage = state_storage
        self._attr_name: Optional[str] = None
        self._transitions: list[StateTransition] = []
        self._enter_state_callbacks: dict[Enum, set[Callable[[object, Enum, Enum], Any]]] = (
            defaultdict(set)
        )
        self._exit_state_callbacks: dict[Enum, set[Callable[[object, Enum, Enum], Any]]] = (
            defaultdict(set)
        )

    def __set_name__(self, owner: type, attr_name: str) -> None:
        """
        Called once at the creation of owner class
        """
        self._attr_name = attr_name

    def __get__(self, instance: object, objtype: Optional[type]) -> Enum:
        """
        Get current state from owner object
        """

        # return descriptor itself when called from class
        if instance is None:
            return self  # type: ignore[return-value]

        return self._get_state(instance)

    def _get_state(self, instance: object) -> Enum:
        if self._state_storage:
            return self._state_storage.get_state(instance)

        else:
            if self._attr_name is None:
                raise ValueError('Cannot get state from unitialized descriptor')

            return getattr(  # type: ignore[return-value]
                instance,
                '_' + self._attr_name,
                self._initial_state,
            )

    def _force_set_state(self, instance: object, state: Enum) -> None:
        current_state = self._get_state(instance)
        if self._state_storage:
            self._state_storage.set_state(instance, state)
        else:
            if self._attr_name is None:
                raise ValueError('Cannot set state via unitialized descriptor')
            setattr(
                instance,
                '_' + self._attr_name,
                state,
            )
        for callback in self._exit_state_callbacks[current_state]:
            callback(instance, current_state, state)
        for callback in self._enter_state_callbacks[state]:
            callback(instance, current_state, state)

    def __set__(self, instance: object, value: Any) -> NoReturn:
        """
        Forbit directly set state
        """
        raise AttributeError('Cannot change state directly. Use transitions')

    def transition(self, source: Union[Enum, Collection[Enum]], dest: Enum) -> StateTransition:
        """
        Create new transition callable
        """
        # check if value is correct Enum
        if dest not in self._all_states:
            raise ValueError('Destination state not found', dest)

        if not isinstance(source, Collection):
            source = [source]
        for source_state in source:
            if source_state not in self._all_states:
                raise ValueError('Source state not found', source)

        transition = StateTransition(
            source, dest, ProxyStateStorage(self._get_state, self._force_set_state)
        )
        self._transitions.append(transition)

        return transition

    def dispatch(self, method: Callable[P, T]) -> StateDispatchedMethod[P, T]:
        dispatcher = StateDispatcher(
            ProxyStateStorage(self._get_state, self._force_set_state), self._all_states, method
        )
        dispatched_method = StateDispatchedMethod(dispatcher)
        return dispatched_method

    def on_transition(
        self, *transitions: StateTransition
    ) -> Callable[[Callable[[object, Enum, Enum], Any]], Callable[[object, Enum, Enum], Any]]:
        def wrapper(
            func: Callable[[object, Enum, Enum], Any],
        ) -> Callable[[object, Enum, Enum], Any]:
            for transition in transitions:
                transition._register_callback(func)
            return func

        if (
            len(transitions) == 1
            and callable(transitions[0])
            and not isinstance(transitions[0], StateTransition)
        ):
            func = transitions[0]
            transitions = self._transitions  # type: ignore[assignment]
            return wrapper(func)

        elif not transitions:
            transitions = self._transitions  # type: ignore[assignment]
        else:
            for transition in transitions:
                if transition not in self._transitions:
                    raise ValueError('Transition not found in current state machine', transition)

        return wrapper

    def on_state_exited(
        self, *states: Enum
    ) -> Callable[[Callable[[object, Enum, Enum], Any]], Callable[[object, Enum, Enum], Any]]:
        def wrapper(
            func: Callable[[object, Enum, Enum], Any],
        ) -> Callable[[object, Enum, Enum], Any]:
            for state in states:
                self._exit_state_callbacks[state].add(func)
            return func

        if len(states) == 1 and callable(states[0]) and not isinstance(states[0], Enum):
            func = states[0]
            states = tuple(self._all_states)
            return wrapper(func)
        else:
            for state in states:
                if state not in self._all_states:
                    raise ValueError('Target state not found', state)

        return wrapper

    def on_state_entered(
        self, *states: Enum
    ) -> Callable[[Callable[[object, Enum, Enum], Any]], Callable[[object, Enum, Enum], Any]]:
        def wrapper(
            func: Callable[[object, Enum, Enum], Any],
        ) -> Callable[[object, Enum, Enum], Any]:
            for state in states:
                self._enter_state_callbacks[state].add(func)
            return func

        if len(states) == 1 and callable(states[0]) and not isinstance(states[0], Enum):
            func = states[0]
            states = tuple(self._all_states)
            return wrapper(func)
        else:
            for state in states:
                if state not in self._all_states:
                    raise ValueError('Target state not found', state)

        return wrapper



from enum import Enum, auto
from typing import Any
from fsmate import StateDescriptor
from dataclasses import dataclass


class DoorState(Enum):
    closed = auto()
    opening = auto()
    opened = auto()
    closing = auto()


@dataclass
class GarageDoor:
    motor: Any
    alarm: Any

    state = StateDescriptor(DoorState, DoorState.closed)

    start_opening = state.transition(DoorState.closed, DoorState.opening)
    finish_opening = state.transition(DoorState.opening, DoorState.opened)
    start_closing = state.transition(DoorState.opened, DoorState.closing)
    finish_closing = state.transition(DoorState.closing, DoorState.closed)

    @state.on_transition(start_opening)
    def on_start_opening(self, _source, _dest):
        self.motor.up()

    @state.on_transition(finish_opening)
    def on_finish_opening(self, _source, _dest):
        self.motor.stop()

    @state.on_transition(start_closing)
    def on_start_closing(self, _source, _dest):
        self.motor.down()
        self.alarm.beep()

    @state.on_transition(finish_closing)
    def on_finish_closing(self, _source, _dest):
        self.motor.stop()

    @state.dispatch
    def push_button(self):
        self.alarm.beep()  # default behavior

    @push_button.overload(DoorState.closed)
    def push_button_closed(self):
        self.start_opening()

    @push_button.overload(DoorState.opened)
    def push_button_opened(self):
        self.start_closing()


class Alarm:
    def beep(self):
        print('beep')


class Motor:
    def up(self):
        print('motor up')

    def down(self):
        print('motor down')

    def stop(self):
        print('motor stop')


door = GarageDoor(Motor(), Alarm())
door.push_button()  # motor up
door.finish_opening()  # motor stop

door.push_button()  # motor down & beep

door.finish_closing()  # motor stop

door.push_button()  # motor up
door.push_button()  # beep
door.push_button()  # beep



from fsmate import StateDescriptor
from enum import Enum, auto
from random import randint


class GameState(Enum):
    Initial = auto()
    WelcomeMessage = auto()
    Difficulty = auto()
    Attempts = auto()
    GuessNumber = auto()
    GameOver = auto()
    ExitMessage = auto()


class Game:
    score: int = 0
    difficulty: int
    max_attempts: int

    state = StateDescriptor(GameState, GameState.Initial)

    welcome = state.transition(GameState.Initial, GameState.WelcomeMessage)

    configure = state.transition(GameState, GameState.Difficulty)
    select_attempts = state.transition(GameState.Difficulty, GameState.Attempts)

    restart_game = state.transition(GameState.GameOver, GameState.GuessNumber)
    start_game = state.transition(GameState.Attempts, GameState.GuessNumber)

    correct_number = state.transition(GameState.GuessNumber, GameState.GuessNumber)
    game_over = state.transition(GameState.GuessNumber, GameState.GameOver)

    exit = state.transition(GameState, GameState.ExitMessage)

    def run(self):
        self.welcome()

        while self.state != GameState.ExitMessage:
            self.process()

    @state.on_transition(exit)
    def print_exit_message(self, from_state, to_state):
        print(f'Total score: {self.score}')
        print('Bye!')

    @state.on_state_entered(GameState.GuessNumber)
    def guess_number(self, from_state, to_state):
        self.number = randint(1, self.difficulty)
        self.attempts_left = self.max_attempts
        print('I made a number. Can you guess it?')

    @state.on_transition(correct_number)
    def next_number(self, from_state, to_state):
        self.score += 1

    @state.on_transition(restart_game)
    def reset_score(self, from_state, to_state):
        print('Score was resetted')
        self.score = 0

    @state.dispatch
    def process(self) -> None:
        raise NotImplementedError()

    @state.on_state_entered(GameState.WelcomeMessage)
    def on_welcome_message_entered(self, from_state, to_state):
        print('Welcome to guess-the-number game!')
        print("Type 'q' to exit")
        print('Type anything to continue')

    @process.overload(GameState.WelcomeMessage)
    def process_welcome_message(self):
        res = input('>>> ')

        if res == 'q':
            self.exit()
        else:
            self.configure()

    @state.on_state_entered(GameState.Difficulty)
    def on_difficulty_entered(self, from_state, to_state):
        print('select difficulty (1-10)')
        print("Type 'q' to exit")

    @process.overload(GameState.Difficulty)
    def process_set_difficulty(self):
        res = input('>>> ')
        if res == 'q':
            self.exit()
            return
        try:
            res = int(res)
        except ValueError:
            return

        if res not in range(1, 101):
            return

        self.difficulty = res
        self.select_attempts()

    @state.on_state_entered(GameState.Attempts)
    def on_attemtps_entered(self, from_state, to_state):
        print('select attempts (1-10)')
        print("Type 'q' to exit")

    @process.overload(GameState.Attempts)
    def process_select_attempts(self):
        res = input('>>> ')
        if res == 'q':
            self.exit()
            return
        try:
            res = int(res)
        except ValueError:
            return
        if res not in range(1, 11):
            return
        self.max_attempts = res
        self.start_game()

    @process.overload(GameState.GuessNumber)
    def process_guess_number(self):
        if self.attempts_left <= 0:
            self.game_over()
            return
        print(f'Score: {self.score}')
        print(f'Attempts left: {self.attempts_left}')
        print('Type a number to guess')
        print("Type 'q' to exit")
        print("Type 'c' to change difficulty (score will be lost)")

        res = input('>>> ')
        if res == 'q':
            self.exit()
            return
        elif res == 'c':
            self.configure()
            return

        try:
            res = int(res)
        except ValueError:
            return

        if res > self.number:
            print('Your number in bigger')
            self.attempts_left -= 1
        elif res < self.number:
            print('Your number is lower')
            self.attempts_left -= 1
        else:
            print('Correct!')
            self.correct_number()

    @state.on_state_entered(GameState.GameOver)
    def on_game_over_entered(self, from_state, to_state):
        print('Game Over!')
        print(f'The number was {self.number}')
        print(f'Final score: {self.score}')
        print("Type 'r' to start again")
        print("Type 'c' to change difficulty")
        print("Type 'q' to exit")

    @process.overload(GameState.GameOver)
    def process_game_over(self):
        res = input('>>> ')
        if res == 'r':
            self.restart_game()
        elif res == 'c':
            self.configure()
        elif res == 'q':
            self.exit()


if __name__ == '__main__':
    game = Game()
    game.run()
